package com.htfp.weather.web.service;

import com.aliyun.tablestore.grid.consts.AttributionEnum;
import com.aliyun.tablestore.grid.model.GridDataSet;
import com.aliyun.tablestore.grid.model.GridDataSetMeta;
import com.aliyun.tablestore.grid.model.grid.Grid4D;
import com.htfp.weather.griddata.common.GfsVariableHeightEnum;
import com.htfp.weather.griddata.common.GfsVariableIsobaricEnum;
import com.htfp.weather.griddata.common.TableConfigBean;
import com.htfp.weather.griddata.operation.GfsDataFetcher;
import com.htfp.weather.griddata.operation.QueryMeta;
import com.htfp.weather.info.Constant;
import com.htfp.weather.utils.DateTimeUtils;
import com.htfp.weather.utils.NdArrayUtils;
import com.htfp.weather.web.exception.AppExcpetion;
import com.htfp.weather.web.exception.ErrorCode;
import com.htfp.weather.web.pojo.response.NowWeatherStatus;
import com.htfp.weather.web.pojo.response.PlaneResponse;
import com.htfp.weather.web.pojo.response.ProfileResponse;
import com.htfp.weather.web.pojo.response.TimeSeriesDataset;
import lombok.extern.slf4j.Slf4j;
import org.springframework.stereotype.Service;
import org.springframework.util.CollectionUtils;
import ucar.ma2.Array;

import javax.annotation.Resource;
import java.time.Duration;
import java.time.Instant;
import java.time.OffsetDateTime;
import java.time.ZoneOffset;
import java.time.format.DateTimeFormatter;
import java.util.*;
import java.util.stream.Collectors;

/**
 * @Author : shiyi
 * @Date : 2024/2/2 15:47
 * @Description : 获取tablestore格点数据
 */
@Service("tablestore-gfs") @Slf4j
public class GfsDataServiceImpl implements IDataService {
    private static final int SURFACE_LEVEL = 9999;
    @Resource
    GfsDataFetcher gfsDataFetcher;
    @Resource
    QueryMeta queryMeta;
    @Resource
    TableConfigBean tableConfigBean;


    public ProfileResponse getProfile(OffsetDateTime targetTime, String variableName, double latitude, double longitude) {
        String targetVariable = GfsVariableIsobaricEnum.getGfsVariableName(variableName);

        if (targetVariable == null) {
            throw new AppExcpetion(ErrorCode.VALIDATE_ERROR, variableName + "变量在数据库中不存在");
        }

        GridDataSetMeta lastGridDataSetMeta = null;
        try {
            lastGridDataSetMeta = queryMeta.getLastGridDataSetMeta();
        } catch (Exception e) {
            throw new AppExcpetion(ErrorCode.DATA_SET_EMPTY, ": e.getMessage()");
        }
        int iTime = getTargetTimeIndex(targetTime, lastGridDataSetMeta);
        int iLat = getLatitudeIndex(latitude);
        int iLon = getLongitudeIndex(longitude);
        try {
            String gridDataSetId = lastGridDataSetMeta.getGridDataSetId();
            int[] pressureLevels = tableConfigBean.getPressureList();
            Array array = gfsDataFetcher.getProfile(gridDataSetId, targetVariable, iTime, iLat, iLon);
            float[] values = (float[]) array.getStorage();
            return new ProfileResponse(pressureLevels, values);
        } catch (Exception e) {
            throw new RuntimeException(e);
        }

    }

    public TimeSeriesDataset getForecastSeries(double latitude, double longitude, int level) {
        GridDataSetMeta lastGridDataSetMeta = null;
        try {
            lastGridDataSetMeta = queryMeta.getLastGridDataSetMeta();
        } catch (Exception e) {
            throw new AppExcpetion(ErrorCode.DATA_SET_EMPTY, ": e.getMessage()");
        }
        Map<String, List> params = getFutureTimeIndexList(lastGridDataSetMeta);
        List<Integer> iTimeList = params.get("iTimeList");
        List<String> timeList = params.get("timeList");
        int[] iTimes = iTimeList.stream().mapToInt(Integer::valueOf).toArray();
        int iLat = getLatitudeIndex(latitude);
        int iLon = getLongitudeIndex(longitude);
        int iLev = getLevelIndex(level);
        List<String> variableNames = getVariableNamesInDatabase(level);

        try {
            String gridDataSetId = lastGridDataSetMeta.getGridDataSetId();
            GridDataSet gridDataSet = gfsDataFetcher.getSeries(gridDataSetId, variableNames, iTimes, iLev, iLat, iLon);
            if (level == SURFACE_LEVEL) {
                return buildTimeSeriesDatasetSurface(timeList, gridDataSet);
            } else {
                return buildTimeSeriesDatasetPressure(timeList, gridDataSet);
            }

        } catch (Exception e) {
            throw new RuntimeException(e);
        }
    }

    public NowWeatherStatus getNowWeather(double latitude, double longitude, int level) {
        GridDataSetMeta lastGridDataSetMeta = null;
        try {
            lastGridDataSetMeta = queryMeta.getLastGridDataSetMeta();
        } catch (Exception e) {
            throw new AppExcpetion(ErrorCode.DATA_SET_EMPTY, ": e.getMessage()");
        }
        int iTime = getTargetTimeIndex(OffsetDateTime.now(), lastGridDataSetMeta);
        int iLat = getLatitudeIndex(latitude);
        int iLon = getLongitudeIndex(longitude);
        int iLev = getLevelIndex(level);
        List<String> variableNames = getVariableNamesInDatabase(level);

        try {
            String gridDataSetId = lastGridDataSetMeta.getGridDataSetId();
            GridDataSet gridDataSet = gfsDataFetcher.getSinglePoint(gridDataSetId, variableNames, iTime, iLev, iLat, iLon);
            if (level == SURFACE_LEVEL) {
                return buildNowWeatherSurface(gridDataSet);
            } else {
                return buildNowWeatherPressure(gridDataSet);
            }

        } catch (Exception e) {
            throw new RuntimeException(e);
        }
    }


    private List<String> getVariableNamesInDatabase(int level) {
        List<String> variableNames = new ArrayList<>();
        if (level == SURFACE_LEVEL) {
            for (String e: tableConfigBean.getVariableList()){
                String gfsVariableName = GfsVariableHeightEnum.getGfsVariableName(e);
                if (gfsVariableName != null){
                    variableNames.add(gfsVariableName);
                }
            }
        } else {
            for (String e: tableConfigBean.getVariableList()){
                String gfsVariableName = GfsVariableIsobaricEnum.getGfsVariableName(e);
                if (gfsVariableName != null){
                    variableNames.add(gfsVariableName);
                }
            }
        }
        return variableNames;
    }

    public PlaneResponse getPlane(OffsetDateTime targetTime, String variableName, int level, double minLat, double maxLat, double minLon, double maxLon) {
        int iLev;
        String targetVariable;
        if (level == SURFACE_LEVEL) {
            iLev = 0; // NOTE 2024/5/20: 0 表示地面，风速对应10m高度，其他变量对应2m高度
            targetVariable = GfsVariableHeightEnum.getGfsVariableName(variableName);
        } else {
            iLev = getPressureIndex(level);
            targetVariable = GfsVariableIsobaricEnum.getGfsVariableName(variableName);
        }

        if (targetVariable == null) {
            throw new AppExcpetion(ErrorCode.VALIDATE_ERROR, variableName + "变量在数据库中不存在");
        }
        GridDataSetMeta lastGridDataSetMeta = null;
        try {
            lastGridDataSetMeta = queryMeta.getLastGridDataSetMeta();
        } catch (Exception e) {
            throw new AppExcpetion(ErrorCode.DATA_SET_EMPTY, ": e.getMessage()");
        }

        int iTime = getTargetTimeIndex(targetTime, lastGridDataSetMeta);
        int iMinLat = getLatitudeIndex(minLat);
        int iMaxLat = getLatitudeIndex(maxLat);
        int iMinLon = getLongitudeIndex(minLon);
        int iMaxLon = getLongitudeIndex(maxLon);
        int latLength = iMaxLat - iMinLat + 1;
        int lonLength = iMaxLon - iMinLon + 1;

        try {
            String gridDataSetId = lastGridDataSetMeta.getGridDataSetId();
            Array array = gfsDataFetcher.getPlane(gridDataSetId, targetVariable, iTime, iLev, iMinLat, iMinLon, latLength, lonLength);

            return buildPlane(array, iMinLat, iMinLon, latLength, lonLength);
        } catch (Exception e) {
            throw new RuntimeException(e);
        }

    }

    private PlaneResponse buildPlane(Array array, int iMinLat, int iMinLon, int latLength, int lonLength) {
        PlaneResponse planeResponse = new PlaneResponse();
        float[][] values = (float[][]) array.reduce().copyToNDJavaArray();
        planeResponse.setLatList(Arrays.copyOfRange(tableConfigBean.getLatList(), iMinLat, iMinLat + latLength));
        planeResponse.setLonList(Arrays.copyOfRange(tableConfigBean.getLonList(), iMinLon, iMinLon + lonLength));
        planeResponse.setValues(values);
        return planeResponse;
    }

    private TimeSeriesDataset buildTimeSeriesDatasetPressure(List<String> timeList, GridDataSet gridDataSet) {
        TimeSeriesDataset timeSeriesDataset = new TimeSeriesDataset(timeList.size());
        timeSeriesDataset.setTime(timeList);
        Grid4D grid4D;
        grid4D = gridDataSet.getVariable(GfsVariableIsobaricEnum.getGfsVariableName(GfsVariableIsobaricEnum.TEMP));
        if (grid4D != null) {
            timeSeriesDataset.setTemp((float[]) grid4D.toArray().getStorage());
        }
        grid4D = gridDataSet.getVariable(GfsVariableIsobaricEnum.getGfsVariableName(GfsVariableIsobaricEnum.WIND_SPEED));
        if (grid4D != null) {
            timeSeriesDataset.setWindSpeed((float[]) grid4D.toArray().getStorage());
        }
        grid4D = gridDataSet.getVariable(GfsVariableIsobaricEnum.getGfsVariableName(GfsVariableIsobaricEnum.WIND360));
        if (grid4D != null) {
            timeSeriesDataset.setWind360((float[]) grid4D.toArray().getStorage());
        }
        grid4D = gridDataSet.getVariable(GfsVariableIsobaricEnum.getGfsVariableName(GfsVariableIsobaricEnum.CLOUD));
        if (grid4D != null) {
            timeSeriesDataset.setCloud((float[]) grid4D.toArray().getStorage());
        }
        return timeSeriesDataset;
    }

    private TimeSeriesDataset buildTimeSeriesDatasetSurface(List<String> timeList, GridDataSet gridDataSet) {
        TimeSeriesDataset timeSeriesDataset = new TimeSeriesDataset(timeList.size());
        timeSeriesDataset.setTime(timeList);
        Grid4D grid4D;
        grid4D = gridDataSet.getVariable(GfsVariableHeightEnum.getGfsVariableName(GfsVariableHeightEnum.TEMP));
        if (grid4D != null) {
            timeSeriesDataset.setTemp((float[]) grid4D.toArray().getStorage());
        }
        grid4D = gridDataSet.getVariable(GfsVariableHeightEnum.getGfsVariableName(GfsVariableHeightEnum.WIND_SPEED));
        if (grid4D != null) {
            timeSeriesDataset.setWindSpeed((float[]) grid4D.toArray().getStorage());
        }
        grid4D = gridDataSet.getVariable(GfsVariableHeightEnum.getGfsVariableName(GfsVariableHeightEnum.WIND360));
        if (grid4D != null) {
            timeSeriesDataset.setWind360((float[]) grid4D.toArray().getStorage());
        }
        return timeSeriesDataset;
    }

    private NowWeatherStatus buildNowWeatherPressure(GridDataSet gridDataSet) {
        String timeStr = OffsetDateTime.now().withMinute(0).withNano(0).format(DateTimeFormatter.ofPattern(Constant.API_TIME_STRING));
        NowWeatherStatus nowWeatherStatus = new NowWeatherStatus();
        nowWeatherStatus.setTime(timeStr);
        Grid4D grid4D;
        grid4D = gridDataSet.getVariable(GfsVariableIsobaricEnum.getGfsVariableName(GfsVariableIsobaricEnum.TEMP));
        if (grid4D != null) {
            nowWeatherStatus.setTemp(((float[]) grid4D.toArray().getStorage())[0]);
        }
        grid4D = gridDataSet.getVariable(GfsVariableIsobaricEnum.getGfsVariableName(GfsVariableIsobaricEnum.WIND_SPEED));
        if (grid4D != null) {
            nowWeatherStatus.setWindSpeed(((float[]) grid4D.toArray().getStorage())[0]);
        }
        grid4D = gridDataSet.getVariable(GfsVariableIsobaricEnum.getGfsVariableName(GfsVariableIsobaricEnum.WIND360));
        if (grid4D != null) {
            nowWeatherStatus.setWind360(((float[]) grid4D.toArray().getStorage())[0]);
        }
        grid4D = gridDataSet.getVariable(GfsVariableIsobaricEnum.getGfsVariableName(GfsVariableIsobaricEnum.CLOUD));
        if (grid4D != null) {
            nowWeatherStatus.setCloud(((float[]) grid4D.toArray().getStorage())[0]);
        }
        return nowWeatherStatus;
    }

    private NowWeatherStatus buildNowWeatherSurface(GridDataSet gridDataSet) {
        String timeStr = OffsetDateTime.now().withMinute(0).withNano(0).format(DateTimeFormatter.ofPattern(Constant.API_TIME_STRING));
        NowWeatherStatus nowWeatherStatus = new NowWeatherStatus();
        nowWeatherStatus.setTime(timeStr);
        Grid4D grid4D;
        grid4D = gridDataSet.getVariable(GfsVariableHeightEnum.getGfsVariableName(GfsVariableHeightEnum.TEMP));
        if (grid4D != null) {
            nowWeatherStatus.setTemp(((float[]) grid4D.toArray().getStorage())[0]);
        }
        grid4D = gridDataSet.getVariable(GfsVariableHeightEnum.getGfsVariableName(GfsVariableHeightEnum.WIND_SPEED));
        if (grid4D != null) {
            nowWeatherStatus.setWindSpeed(((float[]) grid4D.toArray().getStorage())[0]);
        }
        grid4D = gridDataSet.getVariable(GfsVariableHeightEnum.getGfsVariableName(GfsVariableHeightEnum.WIND360));
        if (grid4D != null) {
            nowWeatherStatus.setWind360(((float[]) grid4D.toArray().getStorage())[0]);
        }
        return nowWeatherStatus;
    }
    private int getTargetTimeIndex(OffsetDateTime targetTime, GridDataSetMeta lastGridDataSetMeta) {
        long milli = (long) lastGridDataSetMeta.getAttributes().get(AttributionEnum.REFERENCE_TIME.getName());
        OffsetDateTime refTime = OffsetDateTime.ofInstant(Instant.ofEpochMilli(milli), ZoneOffset.ofHours(0));
        if (targetTime.isBefore(refTime)) {
            throw new AppExcpetion(ErrorCode.QUERY_TIME_ERROR);
        }

        List<String> forecastHours = lastGridDataSetMeta.getForecastHours();
        String targetHour = String.valueOf(getForecastHourFromTargetTime(targetTime, refTime));
        if (!forecastHours.contains(targetHour)) {
            throw new AppExcpetion(ErrorCode.QUERY_TIME_ERROR);
        }
        int iTime = forecastHours.indexOf(targetHour);
        return iTime;
    }

    private Map<String, List> getFutureTimeIndexList(GridDataSetMeta lastGridDataSetMeta) {
        long milli = (long) lastGridDataSetMeta.getAttributes().get(AttributionEnum.REFERENCE_TIME.getName());
        OffsetDateTime refTime = OffsetDateTime.ofInstant(Instant.ofEpochMilli(milli), ZoneOffset.ofHours(0));
        OffsetDateTime currentTime = DateTimeUtils.getUTCDateTime(OffsetDateTime.now());
        if (currentTime.isBefore(refTime)) {
            throw new AppExcpetion(ErrorCode.QUERY_TIME_ERROR);
        }
        int startHour = getForecastHourFromTargetTime(currentTime, refTime);
        List<String> forecastHours = lastGridDataSetMeta.getForecastHours();
        List<Integer> iTimeList = new ArrayList<>();
        List<String> timeList = new ArrayList<>();

        for (int i = 0; i < forecastHours.size(); i++) {
            int hour = Integer.parseInt(forecastHours.get(i));
            if (hour >= startHour) {
                iTimeList.add(i);
                String timeStr = DateTimeUtils.getLocalZoneDateTime(refTime.plusHours(hour))
                        .format(DateTimeFormatter.ofPattern(Constant.API_TIME_STRING));
                timeList.add(timeStr);
            }
            if (iTimeList.size() >= 24) {
                break;
            }
        }
        if (CollectionUtils.isEmpty(iTimeList) || CollectionUtils.isEmpty(timeList)) {
            throw new AppExcpetion(ErrorCode.QUERY_TIME_ERROR);
        }
        Map<String, List> map = new HashMap<>(2);
        map.put("iTimeList", iTimeList);
        map.put("timeList", timeList);
        return map;
    }
    private int getLongitudeIndex(double lon) {
        double[] lonList = tableConfigBean.getLonList();
        if (lonList.length == 0) {
            throw new AppExcpetion(ErrorCode.LONGITUDE_INDEX_ERROR, "经度列表为空， 无法查找索引");
        }
        double minLon = lonList[0];
        double maxLon = lonList[lonList.length - 1];
        if (minLon > lon || lon > maxLon) {
            throw new AppExcpetion(ErrorCode.LONGITUDE_INDEX_ERROR, "经度不在有效范围内: " + minLon + " ~ " + maxLon);
        }
        int nearestIndex = NdArrayUtils.findNearestIndex(lonList, lon);
        if (nearestIndex == -1) {
            throw new AppExcpetion(ErrorCode.LONGITUDE_INDEX_ERROR, "找不到" + lon + "的经度索引");
        }
        return nearestIndex;
    }

    private int getLatitudeIndex(double lat) {
        double[] latList = tableConfigBean.getLatList();
        if (latList.length == 0) {
            throw new AppExcpetion(ErrorCode.LATITUDE_INDEX_ERROR, "纬度列表为空， 无法查找索引");
        }
        double minLat = latList[0];
        double maxLat = latList[latList.length - 1];

        if (minLat > lat || lat > maxLat) {
            throw new AppExcpetion(ErrorCode.LATITUDE_INDEX_ERROR, "纬度不在有效范围内: " + minLat + " ~ " + maxLat);
        }
        int nearestIndex = NdArrayUtils.findNearestIndex(latList, lat);
        if (nearestIndex == -1) {
            throw new AppExcpetion(ErrorCode.LATITUDE_INDEX_ERROR, "找不到" + lat + "的纬度索引");
        }
        return nearestIndex;
    }

    private int getPressureIndex(int pressureLevel) {
        List<Integer> levList = Arrays.stream(tableConfigBean.getPressureList()).boxed().collect(Collectors.toList());
        if (CollectionUtils.isEmpty(levList)) {
            throw new AppExcpetion(ErrorCode.LEVEL_INDEX_ERROR, "气压列表为空， 无法查找索引");
        }
        int index = levList.indexOf(pressureLevel);
        if (index == -1) {
            throw new AppExcpetion(ErrorCode.LATITUDE_INDEX_ERROR, "找不到" + pressureLevel + "的气压高度索引");
        }
        return index;
    }

    private int getLevelIndex(int level) {
        if (level == SURFACE_LEVEL) {
            return 0; // NOTE 2024/5/20: 0 表示地面，风速对应10m高度，其他变量对应2m高度
        } else {
            return getPressureIndex(level);
        }
    }

    private int getForecastHourFromTargetTime(OffsetDateTime targetTime, OffsetDateTime refTime) {
        return (int) Duration.between(refTime, targetTime).toHours();
    }


}
